##############################################################################
#
# Copyright (c) 2010 ViFiB SARL and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################


import unittest
import tempfile
import urllib2
import slapos.libconnection
from slapos.libnetworkcache import NetworkcacheClient
from slapos.tests.libnetworkcachemixin import LibNetworkCacheMixin, start_server, stop_server


class OfflineTest(unittest.TestCase):
  def test_download_offline(self):
    nc = NetworkcacheClient('http://127.0.0.1:0', 'http://127.0.0.1:0')
    self.assertRaises(IOError, nc.download, 'sha512sum')

  def test_upload_offline(self):
    content = tempfile.TemporaryFile()
    nc = NetworkcacheClient('http://127.0.0.1:0', 'http://127.0.0.1:0')
    self.assertRaises(IOError, nc.upload, content)


class TestNetworkcacheClient(LibNetworkCacheMixin):
  """
    Class to test the NetworkcacheClient implementation.
  """

  def setUp(self):
    """ Setup the test. """
    LibNetworkCacheMixin.setUp(self)
    self.host = 'localhost'
    self.port = 8000
    self.shacache_path = '/shacache'
    self.shadir_path = '/shadir'
    url = 'http://%s:%s' % (self.host, self.port)
    self.shacache_url = url + self.shacache_path
    self.shadir_url = url + self.shadir_path
    self.nc = NetworkcacheClient(shacache=self.shacache_url,
                                 shadir=self.shadir_url)

  def test_init_method_normal_http_url(self):
    """
      Check if the init method is setting the attributes correctly.
    """
    self.assertEquals({'Content-Type': 'application/json'}, \
                          self.nc.shacache_header_dict)
    self.assertEquals(self.host, self.nc.shacache_host)
    self.assertEquals(self.shacache_path, self.nc.shacache_path)
    self.assertEquals(self.port, self.nc.shacache_port)
    self.assertEquals(self.shacache_url, self.nc.shacache_url)

    self.assertEquals({'Content-Type': 'application/json'}, \
                         self.nc.shadir_header_dict)
    self.assertEquals(self.host, self.nc.shadir_host)
    self.assertEquals(self.shadir_path, self.nc.shadir_path)
    self.assertEquals(self.port, self.nc.shadir_port)

  def test_signature_creation_without_private_key_file(self):
    """
      Without the private key file, it is not possible to create the
      signature so it must retun an empty string.
    """
    self.assertEquals('', self.nc._getSignatureString())

  def test_signature_creation_with_private_key_file(self):
    """
      Check if the signature creation does not have any error.
    """
    nc = NetworkcacheClient(
                   shacache=self.shacache_url,
                   shadir=self.shadir_url,
                   signature_private_key_file=self.signature_private_key_file)
    self.assertNotEquals('', nc._getSignatureString())

  def test_verification_without_signature_certificate_file_list(self):
    """
      Without the signature certificate file list it is not possible to
      verify if the signature is trusted or not.
      So, the _verifySignatureInCertificateList should return False.
    """
    nc = NetworkcacheClient(
             shacache=self.shacache_url,
             shadir=self.shadir_url,
             signature_private_key_file=self.signature_private_key_file)
    signature_string = nc._getSignatureString()
    self.assertFalse(nc._verifySignatureInCertificateList(signature_string))

  def test_verification_with_signature_certificate_file_list(self):
    """
      With the signature certificate file list it is possible to
      verify if the signature if trusted or not.

      So, the _verifySignatureInCertificateList should return True
      if the signature_string is valid and it should return False if the
      signature_string is not correct.
    """
    nc = NetworkcacheClient(
             shacache=self.shacache_url,
             shadir=self.shadir_url,
             signature_private_key_file=self.signature_private_key_file,
             signature_certificate_file_list=[self.signature_certificate_file])
    signature_string = nc._getSignatureString()
    self.assertTrue(nc._verifySignatureInCertificateList(signature_string))

    wrong_signature_string = 'InvalidSignatureString'.encode('base64')
    result_bool = nc._verifySignatureInCertificateList(wrong_signature_string)
    self.assertFalse(result_bool)

  def test_verification_with_signature_certificate_file_list_url(self):
    """
      NetworkcacheClient supports to have the certification file under an HTTP
      server.

      During the _verifySignatureInCertificateList method, it'll try to
      download the certification from the given URL and check if the signature
      is valid.
    """
    nc = NetworkcacheClient(
             shacache=self.shacache_url,
             shadir=self.shadir_url,
             signature_private_key_file=self.signature_private_key_file,
             signature_certificate_file_list=['http://localhost:0/public.pem'])
    signature_string = nc._getSignatureString()
    self.assertRaises(IOError, \
                        nc._verifySignatureInCertificateList, signature_string)

  def test_verification_with_non_valid_remote_https_server(self):
    """
      If the HTTPS server does not has a valid certificated,
      URLError must be raised, because we don't trunk on such server.
    """
    https_server_port, https_server_thread, file_pem = start_server()
    https_server_url = 'https://localhost:%s' % https_server_port
    try:
      nc = NetworkcacheClient(
             shacache=self.shacache_url,
             shadir=self.shadir_url,
             signature_private_key_file=self.signature_private_key_file,
             signature_certificate_file_list=[https_server_url + self.signature_certificate_file])
      signature_string = nc._getSignatureString()
      self.assertRaises(urllib2.URLError,
                       nc._verifySignatureInCertificateList, signature_string)
    finally:
      stop_server(https_server_url, https_server_thread)

  def test_verification_with_valid_remote_https_server(self):
    """
      If the HTTPS server is a trustable server we must download the
      certificate, without any problem.
    """
    https_server_port, https_server_thread, file_pem = start_server()
    https_server_url = 'https://localhost:%s' % https_server_port

    # making the https server a valid server
    slapos.libconnection.CERTIFICATE_FILE_LOCATION_LIST = [file_pem]

    https_cert_url = https_server_url + self.signature_certificate_file
    try:
      nc = NetworkcacheClient(
             shacache=self.shacache_url,
             shadir=self.shadir_url,
             signature_private_key_file=self.signature_private_key_file,
             signature_certificate_file_list=[https_cert_url])
      signature_string = nc._getSignatureString()
      nc._verifySignatureInCertificateList(signature_string)
    finally:
      stop_server(https_server_url, https_server_thread)

  def test_signature_verification_priority(self):
    """
      During the signature vefirication, the filesystem path has priority over
      urls. It will only download the certificate from the url if the local
      certificates are not valid.
    """
    nc = NetworkcacheClient(
           shacache=self.shacache_url,
           shadir=self.shadir_url,
           signature_private_key_file=self.signature_private_key_file,
           signature_certificate_file_list=['http://localhost:0/public.pem',
                                            self.signature_certificate_file])
    signature_string = nc._getSignatureString()
    self.assertTrue(nc._verifySignatureInCertificateList(signature_string))
